; See http://rubiksim.sourceforge.net/

#include-once

#include "Utilities.au3"

Global $fRotateCube = False
Global $fRotate = 0

Global Const $dRotationScale = 135

Global $ptMousePos[2]
Global $vLastMouse[3]

; Handles mouse and keyboard rotations
Func Rotate( ByRef $fRotate, $fCubes = False, $fBuild = False )
	If $fRotate = 1 Then
		; Mouse rotations
		RotateCube( $ptMousePos )
		$fSetHomePos = True
	ElseIf $fRotate > 1 Then
		; Keyboard rotations
		If $fCubes And $fInitCubesStatusBarInfo And $fRotate < 7 Then _
			InitCubesStatusBarInfo()
		If $fSetHomePos Then
			$iAngleX = 30
			$iAngleY = 45
			$mRotation = $IDENTITY_MATRIX
			$fSetHomePos = False
		Else
			Switch $fRotate
				Case 2 ; Left 90 degrees
					If $fBuild Then
						$iAngleY -= 90
					Else
						While $iAngleY > $iNewAngleY
							$iAngleY -= 5
							DisplayCube( $iSize, _
								$iAngleX, $iAngleY )
							Sleep(10)
						WEnd
					EndIf
					If $iAngleY < 0 Then _
						$iAngleY += 360

				Case 3 ; Right 90 degrees
					If $fBuild Then
						$iAngleY += 90
					Else
						While $iAngleY < $iNewAngleY
							$iAngleY += 5
							DisplayCube( $iSize, _
								$iAngleX, $iAngleY )
							Sleep(10)
						WEnd
					EndIf
					If $iAngleY > 360 Then _
						$iAngleY -= 360

				Case 4 ; Up 60 degrees
					If $fBuild Then
						$iAngleX -= 60
					Else
						While $iAngleX > $iNewAngleX
							$iAngleX -= 5
							DisplayCube( $iSize, _
								$iAngleX, $iAngleY )
							Sleep(10)
						WEnd
					EndIf
					If $iAngleX < 0 Then _
						$iAnglex += 360

				Case 5 ; Down 60 degrees
					If $fBuild Then
						$iAngleX += 60
					Else
						While $iAngleX < $iNewAngleX
							$iAngleX += 5
							DisplayCube( $iSize, _
								$iAngleX, $iAngleY )
							Sleep(10)
						WEnd
					EndIf
					If $iAngleX > 360 Then _
						$iAngleX -= 360

				Case 6 ; Home position
					$iAngleX = $iNewAngleX
					$iAngleY = $iNewAngleY

				Case 7 ; Numeric keys 1-6 or keys 1-6
					$iAngleX = $iNewAngleX
					$iAngleY = $iNewAngleY
					$fSetHomePos = True
					If $fCubes Then _
						$fInitCubesStatusBarInfo = True
			EndSwitch
		EndIf
	EndIf
	$fRotate = 0
EndFunc

Func RotateCubeButtonDown( ByRef $ptMouse )
	$vLastMouse = MapScreenPointToSphere( $ptMouse, $iWidth, $iHeight )
	$fRotateCube = True
EndFunc

Func RotateCubeButtonUp()
	$fRotateCube = False
EndFunc

Func RotateCube( ByRef $ptMouse )
	Local $vMouse = MapScreenPointToSphere( $ptMouse, $iWidth, $iHeight )
	CalcRotation( $vMouse )
	$vLastMouse = $vMouse
EndFunc

; Map a point on the screen to a 3D location on a sphere
Func MapScreenPointToSphere( ByRef $pt, $w, $h )
	Local $v[3] = [ _
		( 2 * $pt[0] - $w ) / $w, _
		( $h - 2 * $pt[1] ) / $h, _
		0 ]

	Local $d = VectorLength3( $v )
	If $d > 1 Then $d = 1
	$v[2] = Sqrt( 1.001 - $d * $d )	; Project the line segment up to the surface of the sphere

	VectorNormalize3( $v )
	Return $v
EndFunc

Func CalcRotation( ByRef $vMouse )
	Local Static $iRotations = 0
	Local Static $iMaxRotations = 100

	Local $vDirection[3] = [ _
		$vMouse[0] - $vLastMouse[0], _
		$vMouse[1] - $vLastMouse[1], _
		$vMouse[2] - $vLastMouse[2] ]

	Local $dVelocity = VectorLength3( $vDirection )
	If $dVelocity < 0.0001 Then Return

	; Rotate about the axis that is perpendicular to the circle connecting the mouse movements
	Local $RotationAxis = VectorCrossProduct( $vLastMouse, $vMouse )
	Local $RotationAngle = $dVelocity * $dRotationScale

	glPushMatrix()
		glLoadIdentity()
		glRotatef( $RotationAngle, $RotationAxis[0], $RotationAxis[1], $RotationAxis[2] )
		glMultMatrixf( $mRotation )
		glGetFloatv( $GL_MODELVIEW_MATRIX, $mRotation )
	glPopMatrix()

	; Because of accumulated rounding errors in the rotation matrix
	; perform a normalization and "orthogonalization" from time to time.
	$iRotations += 1
	If $iRotations > $iMaxRotations Then
		NormalizeRotationMatrix()
		$iRotations = 0
	EndIf
EndFunc

Func NormalizeRotationMatrix()
	; Get two vectors from the first two rows of the rotation matrix
	; See the note about OpenGL matrices in OGLconsts.au3
	Local $v1[3] = [ $mRotation[0], $mRotation[4], $mRotation[8] ] ; Row 1
	Local $v2[3] = [ $mRotation[1], $mRotation[5], $mRotation[9] ] ; Row 2

	VectorNormalize3( $v1 )
	VectorNormalize3( $v2 )

	; Put normalized row 2 back into matrix
	$mRotation[1] = $v2[0]
	$mRotation[5] = $v2[1]
	$mRotation[9] = $v2[2]

	; Put new "orthogonalized" row 3 back into matrix
	Local $v3 = VectorCrossProduct( $v1, $v2 )
	$mRotation[2]  = $v3[0]
	$mRotation[6]  = $v3[1]
	$mRotation[10] = $v3[2]

	; Put new "orthogonalized" row 1 back into matrix
	$v1 = VectorCrossProduct( $v2, $v3 )
	$mRotation[0] = $v1[0]
	$mRotation[4] = $v1[1]
	$mRotation[8] = $v1[2]
EndFunc
