#include <GUIConstantsEx.au3>
#include <StaticConstants.au3>

#include "APIConstants.au3"

Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

;Global Const $tagRECT = "struct; long Left;long Top;long Right;long Bottom; endstruct"

Global Const $dllGDI32    = DllOpen( "gdi32.dll" )
Global Const $dllGLU32    = DllOpen( "glu32.dll" )
Global Const $dllOpenGL32 = DllOpen( "opengl32.dll" )
Global Const $dllUser32   = DllOpen( "user32.dll" )

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iMenuBarHeight   = 20
Global Const $iStatusBarHeight = 28

Global Const $iGuiWidth  = 600
Global Const $iGuiHeight = 600

Global $hGui, $aAccelKeys[1][2]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth, $iHeight

; Menu variables
Global $fMenuState = 1 ; 1 => $fCubesMenu, 2 => $fScrambleMenu, 3 => $fSolveMenu, 4 => $fBuildMenu
Global $fCubesMenu = 1, $fScrambleMenu = 0, $fSolveMenu = 0, $fBuildMenu = 0
Global $fCubesState = 1, $fScrambleState = 0, $fSolveState = 0, $fBuildState = 0

; Menu bar
Global $idPiece, $idPocket, $idRubik, $idRevenge, $idProfessor, $idVcube6, $idVcube7	; Cubes menu ID's
Global $idCube, $idXlayer, $idYlayer, $idZlayer																				; Build menu ID's
Global $idScramble																																		; Scramble menu ID
Global $idSolve																																				; Solve menu ID
Global $idLogsLists, $idLogsColors																										; Logs menu ID
Global $idHome, $idLeft, $idRight, $idUp, $idDown, $id1, $id2, $id3, $id4, $id5, $id6	; Rotate menu ID's
Global $idZoomIn, $idZoomOut																													; Zoom menu ID's
Global $idOptions																																			; Options menu ID
Global $idHelp																																				; Help menu ID

; StatusBar controls
Global $hCubes, $cbName, $cbLabel, $cbNew, $cbUndo, $cbRedo, _		; Window, controls and variables for the Cubes menu
	$nUndo = 0, $idxUndo = 0, $nRedo = 0, $idxRedo = 0, _
	$aUndoLog[100][3], _	; $fXYZlayer, $nCubeLayer, $fClockWise
	$aRedoLog[100][3], _	; $fXYZlayer, $nCubeLayer, $fClockWise
	$fUndoRedoState = 2, $fInitCubesStatusBarInfo = False
Global $hBuild, $bldFirst, $bldPrev, $bldNext, $bldLast, $bldPlay	; Window and buttons for the Build menu
Global $hScramble, $srmNew, $srmNext, $srmPause, $srmPlay, _			; Window, controls and variables for the Scramble menu
	$srmInput
Global $hSolve, $slvRedo, $slvRedoNext, $slvUndo, $slvUndoNext, _	; Window, buttons and variables for the Solve menu
	$slvPause, $slvButtons
Global $hProgress = 0, $idProgress, $proLabel											; Window, control and variable for the ProgressBar

; OpenGL includes
#include "OGLconsts.au3"
#include "OGLfuncs.au3"

; Project includes
#include "Cubes.au3"
#include "Build.au3"
#include "TurnLayer.au3"
#include "Scramble.au3"
#include "SolveCube.au3"
#include "RotateCube.au3"
#include "Utilities.au3"
#include "MenuFncs.au3"
#include "MenuWins.au3"
#include "Keyboard.au3"

MainScript()


#cs##############
#   Main loop   #
#ce##############

; The primary task of this loop is to execute OpenGL drawing functions and to perform menu changes. Activating
; a menu will interrupt the loop. The interruption can easily be done in the middle of an OpenGL drawing cycle.
; Because of this menu changes affecting OpenGL drawing functions should be performed between two drawing cycles.

; There are two levels in the main loop. The OUTER loop is used to switch between OpenGL related menus. The INNER
; loops, one for each menu, are used to execute OpenGL drawing functions and to handle events from the controls in
; the statusbar.

While 1

	If $fMenuState = $fCubesMenu Then

		; --- Cubes menu ---

		$fTurnLayerDisabled = False							; In Cubes mode it's allowed to rotate a layer in the cube with the mouse. And this is only
																						; allowed in Cubes mode.
		While 1

			; The two drawing functions DisplayCube() and DisplayTurn() do the
			; actual drawing of the cube by executing precompiled display lists.

			If $iSize = $iNewSize Then
				If $fTurn Then
					DisplayTurn( $iSize, _						; DisplayTurn() simulates the rotation of a layer by turning the layer some few degrees at a time.
						$iAngleX, $iAngleY, _						; This is a pure OpenGL rotation. It will certainly rotate the layer 90 degrees in a proper way.
						$nDeltaDegrees )
					If $fTurnDone Then								; When the rotation is done $fTurnDone is set to one. The TurnLayer() function does the actual
						If $fUndoRedoState = 0 Then _		; rotation of the pieces in the cube structure by changing the colors of the sides in the pieces
							$fUndoRedoState = 1						; and by reposition the pieces in the layer. If there is an error in this function, you'll see a
						$fTurn = False									; sudden change of colors for some of the pieces when the rotation stops.
						TurnLayer()
					EndIf
				Else
					DisplayCube( $iSize, _						; DisplayCube() draws the cube when there are no rotations
						$iAngleX, $iAngleY )						
					If $fTurnDone = 1 Then _					; It's important that the next rotation isn't started before the previous one is finished. This
						$fTurnDone = 2									; is controlled by $fTurnDone.
					If $fUndoRedoState = 1 Then _			; It's important that the next Undo/Redo isn't started before the previous one is finished. This
						$fUndoRedoState = 2							; is controlled by $fUndoRedoState.
				EndIf

			; Cubes menu change
			ElseIf $iSize <> $iNewSize Then
				If $fTurn Then $fTurn = False				; If a new cube is selected in the middle af a layer rotation then stop the rotation.
				If $iNewSize = -1 Then							; It's enough to stop the rotation because we'll start all over with a completely new cube.
					$iNewSize = $iSize
				Else
					$dScale = 1.0
				EndIf
				NewCube( $iNewSize )
				If $fLogsColorsWin Then _
					RefreshLogsColorsWin()
			EndIf

			If $fCubesState = 2 Then

				; Leaving the Cubes menu

				If $fTurn Then

					; If leaving the menu in the middle af a layer rotation then finish the rotation

					DisplayTurn( $iSize, _						; This code will probably never be executed. In the Cubes menu the layers are rotated manually
						$iAngleX, $iAngleY, _						; with the mouse. There is not much time to select a menu before the rotation is finished.
						$nDeltaDegrees )
					If $fTurnDone Then
						$fTurn = False
						TurnLayer()
						DisplayCube( $iSize, _
							$iAngleX, $iAngleY )
					EndIf

				Else

					; The last layer rotation is finished

					; Select a new menu for the outer loop
					If $fBuildState = 1 Then
						$fMenuState = $fBuildMenu
					ElseIf $fScrambleState = 1 Then
						$fMenuState = $fScrambleMenu
					ElseIf $fSolveState = 1 Then
						$fMenuState = $fSolveMenu
					EndIf

					; Set the Cubes menu to inactive
					$fCubesState = 0

					; Exit the inner loop
					ExitLoop

				EndIf

			EndIf

			If $fRotate Then _
				Rotate( $fRotate, True )

			If $fZoom Then
				$dScale = $dNewScale
				$fZoom = 0
			EndIf

			Sleep(10)

		WEnd

		$fTurnLayerDisabled = True							; Outside the Cubes menu it's not allowed to rotate a layer in the cube with the mouse

	ElseIf $fMenuState = $fScrambleMenu Then

		; --- Scramble menu ---

		; $fScrambleState values:
		;   0 => Scramble menu inactive
		;   1 => Starts a new scrambling
		;   2 => Continue scrambling
		;   3 => Scrambling is finished because the number of scramble steps is reached
		;        $fScrambleState is set to 3 in the bottom of ScrambleCube() in Scramble.au3
		;        When $fScrambleState = 3 the cube will be redrawn once for each loop
		;        Setting $fScrambleState = 3 can be used to idle around without scrambling
		;   4 => A control in the statusbar is activated, see below
		;   5 => A control in the statusbar is activated, see below
		;   6 => Leaving the Scramble menu

		While 1

			If $fTurn Then
				If $fSimulateLayerRotations Then
					DisplayTurn( $iSize, _
						$iAngleX, $iAngleY, 10 )
					If $fTurnDone Then
						$fTurn = False
						TurnLayer()
						If $fScrambleState = 2 Then _
							ScrambleCube()
					EndIf
				Else
					$fTurnDone = 2
					$fTurn = False
					TurnLayer()
					DisplayCube( $iSize, _
						$iAngleX, $iAngleY )
					If $fScrambleState = 2 Then _
						ScrambleCube()
				EndIf
			Else
				DisplayCube( $iSize, _
					$iAngleX, $iAngleY )
			EndIf

			If $fScrambleState = 2 Then

				; If $fScrambleState = 2 the cube is scrambling and there have not been
				; any interruptions through a menu selection or a statusbar control. In
				; this case skip all the tests below and continue scrambling.

			ElseIf $fScrambleState = 1 Then

				; This starts a new scrambling

				$fScrambleState = 2									; $fScrambleState = 2 means continue scrambling
				NewCube( $iNewSize )
				If $fLogsColorsWin Then _
					RefreshLogsColorsWin()
				$nScrambleMax = $nScrambleSteps
				$nScramble = 0
				ScrambleCube()

			ElseIf $fScrambleState = 4 Then

				; A control in the statusbar is activated
				; The input field is updated or the New button pressed

				$fScrambleState = 3									; $fScrambleState = 3 means just idle around without scrambling.
				If $fTurn Then $fTurn = false				; No reason to finish a rotation in this case. Just terminate it.
				NewCube( $iNewSize )								; This will destroy the previous cube anyway and create a new cube.
				If $fLogsColorsWin Then _
					RefreshLogsColorsWin()
				$nScramble = 0											; $nScramble = 0 is used in the button funtions srmNext() and srmPlay() in MenuFncs.au3 to
																						; test if a scrambling is started or continued. $nScramble = 0 means a scrambling is started.
			ElseIf $fScrambleState = 5 Then

				; A control in the statusbar is activated
				; The Next or Play button is pressed

				$fScrambleState = 2									; $fScrambleState = 2 means continue scrambling
				ScrambleCube()

			ElseIf $fScrambleState = 6 Then

				; Leaving the Scramble menu

				If $fTurn Then

					; If leaving the menu in the middle af a layer rotation then finish the rotation

					DisplayTurn( $iSize, _
						$iAngleX, $iAngleY, 10 )				; Interrupting the scrambling by clicking the Scramble menu will probably happen somewhere
					If $fTurnDone Then								; in the middle af a layer rotation. Most time of the scrambling is spent by turning layers.
						$fTurn = False									; In that case the interrupted rotation should be able to finish.
						TurnLayer()
						DisplayCube( $iSize, _
							$iAngleX, $iAngleY )
					EndIf

				Else

					; The last layer rotation is finished

					; Select a new menu for the outer loop
					If $fCubesState = 1 Then
						$fMenuState = $fCubesMenu
					ElseIf $fBuildState = 1 Then
						$fMenuState = $fBuildMenu
					ElseIf $fSolveState = 1 Then
						$fMenuState = $fSolveMenu
					EndIf

					; Set the Scramble menu to inactive
					$fScrambleState = 0

					; Exit the inner loop
					ExitLoop

				EndIf

			EndIf

			If $fRotate Then _
				Rotate( $fRotate )

			If $fZoom Then
				$dScale = $dNewScale
				$fZoom = 0
			EndIf

			Sleep(10)

		WEnd

	ElseIf $fMenuState = $fSolveMenu Then

		; --- Solve menu ---

		; $fSolveState values:
		;   0 => Solve menu inactive
		;   1 => Starts a new solving
		;   2 => Continue solving
		;   3 => Solving is finished because the number of solving steps is reached
		;        $fSolveState is set to 3 in SolveUndo()/SolveRedo() in SolveCube.au3
		;        When $fSolveState = 3 the cube will be redrawn once for each loop
		;        Setting $fSolveState = 3 can be used to idle around without solving
		;   4 => A control in the statusbar is activated, see below
		;   5 => Leaving the Solve menu

		While 1

			If $fTurn Then
				If $fSimulateLayerRotations Then
					DisplayTurn( $iSize, _
						$iAngleX, $iAngleY, 10 )
					If $fTurnDone Then
						$fTurn = False
						TurnLayer()
						If $fSolveState = 2 Then
							If $fSolve = 1 Then
								SolveUndo()
							ElseIf $fSolve = 2 Then
								SolveRedo()
							EndIf
						EndIf
					EndIf
				Else
					$fTurnDone = 2
					$fTurn = False
					TurnLayer()
					DisplayCube( $iSize, _
						$iAngleX, $iAngleY )
					If $fSolveState = 2 Then
						If $fSolve = 1 Then
							SolveUndo()
						ElseIf $fSolve = 2 Then
							SolveRedo()
						EndIf
					EndIf
				EndIf
			Else
				DisplayCube( $iSize, _
					$iAngleX, $iAngleY )
			EndIf

			If $fSolveState = 2 Then

				; If $fSolveState = 2 the cube is solving and there have not been any
				; interruptions through a menu selection or a statusbar control. In
				; this case skip all the tests below and continue solving.

			ElseIf $fSolveState = 1 Then

				; This starts a new solving

				If $fSolve = 1 Then
					If $idxUndo Then
						$fSolveState = 2								; $fSolveState = 2 means continue solving
						SolveUndo()
					Else
						$fSolveState = 3								; $fSolveState = 3 means idle around without solving
					EndIf
				ElseIf $fSolve = 2 Then
					If $idxRedo Then
						$fSolveState = 2								; $fSolveState = 2 means continue solving
						SolveRedo()
					Else
						$fSolveState = 3								; $fSolveState = 3 means idle around without solving
					EndIf
				EndIf

			ElseIf $fSolveState = 3 And $slvButtons Then

				slvButtons()
				$slvButtons = False

			ElseIf $fSolveState = 4 Then

				; A control in the statusbar is activated
				; A Next button is pressed

				$fSolveState = 2										; $fSolveState = 2 means continue solving
				If $fSolve = 1 Then
					SolveUndo()
				ElseIf $fSolve = 2 Then
					SolveRedo()
				EndIf

			ElseIf $fSolveState = 5 Then

				; Leaving the Solve menu

				If $fTurn Then

					; If leaving the menu in the middle af a layer rotation then finish the rotation

					DisplayTurn( $iSize, _
						$iAngleX, $iAngleY, 10 )				; Interrupting the solving by clicking the Solve menu will probably happen somewhere in
					If $fTurnDone Then								; the middle af a layer rotation. Most time of the solving is spent by turning layers.
						$fTurn = False									; In that case the interrupted rotation should be able to finish.
						TurnLayer()
						DisplayCube( $iSize, _
							$iAngleX, $iAngleY )
					EndIf

				Else

					; The last layer rotation is finished

					; Select a new menu for the outer loop
					If $fCubesState = 1 Then
						$fMenuState = $fCubesMenu
					ElseIf $fBuildState = 1 Then
						$fMenuState = $fBuildMenu
					ElseIf $fScrambleState = 1 Then
						$fMenuState = $fScrambleMenu
					EndIf

					; Set the Solve menu to inactive
					$fSolveState = 0

					; Exit the inner loop
					ExitLoop

				EndIf

			EndIf

			If $fRotate Then _
				Rotate( $fRotate )

			If $fZoom Then
				$dScale = $dNewScale
				$fZoom = 0
			EndIf

			Sleep(10)

		WEnd

	ElseIf $fMenuState = $fBuildMenu Then

		; --- Build menu ---

		While 1

			If $fBuild = $fNewBuild Then
				DisplayBuild( $iSize, _							; DisplayBuild() do the actual drawing of the cube by executing a precompiled display list
					$iAngleX, $iAngleY, $fBuild )

			; Build menu change
			ElseIf $fBuild <> $fNewBuild Then
				NewCube( $iNewSize )
				$fBuild = $fNewBuild
				$fPlayBuild = True
				CalcMaxPieces()
				$nPieces = 1
			EndIf

			If $fBuildState = 2 Then

				; Leaving the Build menu

				$fBuild = 0
				$fNewBuild = 0

				If $fTurn Then $fTurn = False

				; Select a new menu for the outer loop
				If $fCubesState = 1 Then
					$fMenuState = $fCubesMenu
					If $iSize = $iNewSize Then _
						$iNewSize = -1
				ElseIf $fScrambleState = 1 Then
					$fMenuState = $fScrambleMenu
				ElseIf $fSolveState = 1 Then
					$fMenuState = $fSolveMenu
					NewCube( $iNewSize )
				EndIf

				; Clear progressbar
				UpdateProgressBar( 0, 1 )

				; Set the Build menu to inactive
				$fBuildState = 0

				; Exit the inner loop
				ExitLoop

			EndIf

			If $fRotate Then _
				Rotate( $fRotate, False, True )

			If $fZoom Then
				$dScale = $dNewScale
				$fZoom = 0
			EndIf

			Sleep(10)

		WEnd

	EndIf

	Sleep(10)

WEnd


#cs################
#   Main script   #
#ce################

Func MainScript()

	$hGui = GUICreate( "Rubik's Cube", $iGuiWidth, $iGuiWidth, 300, 50, $WS_OVERLAPPEDWINDOW )
	GUISetIcon( "res\RubiksCube.ico" )

	; --- Main menu ---

	CreateMainMenu()

	; --- Accelerator table ---

	CreateAcceleratorTable()

	; --- OpenGL window ---

	$iWidth = $iGuiWidth
	$iHeight = $iGuiHeight - $iMenuBarHeight - $iStatusBarHeight
	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetBkColor( 0xD3D3D3 ) ; D3 = 211
	GUISetState()
	GUISwitch( $hGui )

	; --- StatusBar ---

	; Use label as statusbar
	Local $idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )

	; --- GUI events ---

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	; --- Window messages ---

	GUIRegisterMsg( $WM_LBUTTONDOWN,   "WM_LBUTTONDOWN"   )
	GUIRegisterMsg( $WM_LBUTTONUP,     "WM_LBUTTONUP"     )
	GUIRegisterMsg( $WM_RBUTTONDOWN,   "WM_RBUTTONDOWN"   )
	GUIRegisterMsg( $WM_RBUTTONUP,     "WM_RBUTTONUP"     )
	GUIRegisterMsg( $WM_MOUSEMOVE,     "WM_MOUSEMOVE"     )
	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING"        )
	GUIRegisterMsg( $WM_GETMINMAXINFO, "WM_GETMINMAXINFO" )

	; --- Initializations ---

	ReadValuesFromInifile()

	$nScrambleStepsIni = $aScrambleStepsCubes[$iSize-1]
	$nScrambleStepsNew = $nScrambleStepsIni
	$nScrambleSteps = $nScrambleStepsNew

	ReadColorsFromInifile()

	; --- Prepare OpenGL ---

	EnableOpenGL( $hOpenGL, $hDC, $hRC )

	ResizeOpenGLWindow( $iWidth, $iHeight )

	glClearColor( $aBackGround[0]/255, $aBackGround[1]/255, $aBackGround[2]/255, 1.0 )

	glShadeModel( $GL_FLAT )

	glEnable( $GL_DEPTH_TEST )

	; --- Show GUI ---

	CreateCube( $iSize ) ; Rubik's Cube

	GUISetState()
	ShowCubesControls()

EndFunc


#cs########################
#   GUI event functions   #
#ce########################

Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iMenuBarHeight - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	ResizeOpenGLWindow( $iWidth, $iHeight )
	If $fCubesMenu Then
		WinMove( $hCubes, "", 2, $iHeight + 3 )
	ElseIf $fBuildMenu Then
		WinMove( $hBuild, "", 2, $iHeight + 3 )
		WinMove( $hProgress, "", 282, $iHeight + 3 )
	ElseIf $fScrambleMenu Then
		WinMove( $hScramble, "", 2, $iHeight + 3 )
		WinMove( $hProgress, "", 294, $iHeight + 3 )
	ElseIf $fSolveMenu Then
		WinMove( $hSolve, "", 2, $iHeight + 3 )
		WinMove( $hProgress, "", 282, $iHeight + 3 )
	EndIf
EndFunc

Func GUI_EVENT_CLOSE()
	If Not CheckIniValues() Then						; Values <> $IniValues => Values are updated
		If Not CheckDefValues( False ) Then		; Values <> $IniValues => Save values to inifile
			SaveValuesToInifile()
		Else																	; Values =  $IniValues => Delete values from inifile
			DelValuesFromInifile()
		EndIf
	EndIf
	If Not CheckIniColors() Then						; $aColors <> $aIniColors => Colors are updated
		If Not CheckDefColors( False ) Then		; $aColors <> $aDefColors => Save colors to inifile
			SaveColorsToInifile()
		Else																	; $aColors =  $aDefColors => Delete colors from inifile
			DelColorsFromInifile()
		EndIf
	EndIf
	If $hHelpWin Then CloseHelpWin()
	If $hOptionsWin Then CloseOptionsWin()
	If $hLogsListsWin Then CloseLogsListsWin()
	If $hLogsColorsWin Then CloseLogsColorsWin()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	DllClose( $dllGDI32 )
	DllClose( $dllGLU32 )
	DllClose( $dllOpenGL32 )
	DllClose( $dllUser32 )
	GUIDelete( $hGui )
	Exit
EndFunc


#cs#############################
#   Window message functions   #
#ce#############################

; --- Mouse events ---

Func WM_LBUTTONDOWN( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $hWnd, $iMsg, $iwParam
	Local $ptMouse[2] = [ BitAND( $ilParam, 0xFFFF ), BitShift( $ilParam, 16 ) ]
	If $fSwapMouseButtons Then
		RotateCubeButtonDown( $ptMouse )
	Else
		If $fTurnLayerDisabled Then Return
		TurnLayerButtonDown( $ptMouse )
	EndIf
EndFunc

Func WM_LBUTTONUP( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $hWnd, $iMsg, $iwParam
	If $fSwapMouseButtons Then
		RotateCubeButtonUp()
	Else
		If $fTurnLayerDisabled Then Return
		Local $ptMouse[2] = [ BitAND( $ilParam, 0xFFFF ), BitShift( $ilParam, 16 ) ]
		TurnLayerButtonUp( $ptMouse )
	EndIf
EndFunc

Func WM_RBUTTONDOWN( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $hWnd, $iMsg, $iwParam
	Local $ptMouse[2] = [ BitAND( $ilParam, 0xFFFF ), BitShift( $ilParam, 16 ) ]
	If $fSwapMouseButtons Then
		If $fTurnLayerDisabled Then Return
		TurnLayerButtonDown( $ptMouse )
	Else
		RotateCubeButtonDown( $ptMouse )
	EndIf
EndFunc

Func WM_RBUTTONUP( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $hWnd, $iMsg, $iwParam
	If $fSwapMouseButtons Then
		If $fTurnLayerDisabled Then Return
		Local $ptMouse[2] = [ BitAND( $ilParam, 0xFFFF ), BitShift( $ilParam, 16 ) ]
		TurnLayerButtonUp( $ptMouse )
	Else
		RotateCubeButtonUp()
	EndIf
EndFunc

Func WM_MOUSEMOVE( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $hWnd, $iMsg, $iwParam
	If $fRotateCube Then
		$ptMousePos[0] = BitAND( $ilParam, 0xFFFF )
		$ptMousePos[1] = BitShift( $ilParam, 16 )
		$fRotate = 1
	EndIf
EndFunc

; --- Sizing events ---

Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $x = DllStructGetData( $tRECT, "Left" )
			Local $y = DllStructGetData( $tRECT, "Top" )
			$iWidth = DllStructGetData( $tRECT, "Right" ) - $x - 2 * $iBorderWidth
			$iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iMenuBarHeight - $iStatusBarHeight
			WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
			If $fCubesMenu Then
				WinMove( $hCubes, "", 2, $iHeight + 3 )
			ElseIf $fBuildMenu Then
				WinMove( $hBuild, "", 2, $iHeight + 3 )
				WinMove( $hProgress, "", 282, $iHeight + 3 )
			ElseIf $fScrambleMenu Then
				WinMove( $hScramble, "", 2, $iHeight + 3 )
				WinMove( $hProgress, "", 294, $iHeight + 3 )
			ElseIf $fSolveMenu Then
				WinMove( $hSolve, "", 2, $iHeight + 3 )
				WinMove( $hProgress, "", 282, $iHeight + 3 )
			EndIf
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc

Func WM_GETMINMAXINFO( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tMINMAXINFO = DllStructCreate( "int;int;int;int;int;int;int;int;int;int", $ilParam )
			DllStructSetData( $tMINMAXINFO, 7, $iGuiWidth )		; Min width
			DllStructSetData( $tMINMAXINFO, 8, $iGuiHeight )	; Min height
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc

Func ResizeOpenGLWindow( $w, $h )
	glViewport( 0, 0, $w, $h )

	Local $dLeft   = 0.0 - $dSphereDiameter / 2
	Local $dRight  = 0.0 + $dSphereDiameter / 2
	Local $dBottom = 0.0 - $dSphereDiameter / 2
	Local $dTop    = 0.0 + $dSphereDiameter / 2

	Local $dAspect = $w / $h

	If $dAspect < 1.0 Then
		$dBottom /= $dAspect
		$dTop /= $dAspect
	Else
		$dLeft *= $dAspect
		$dRight *= $dAspect
	EndIf

	glMatrixMode( $GL_PROJECTION )

	glLoadIdentity()

	glOrtho( $dLeft, $dRight, $dBottom, $dTop, $zNear, $zFar )

	glMatrixMode( $GL_MODELVIEW )
EndFunc
