; See http://rubiksim.sourceforge.net/

#include-once

#include "MenuWins.au3"
#include "Utilities.au3"

Global $fTurnLayerDisabled = False

Global $fSwapMouseButtons = False

Global $dMinMouseMovement = 5 ; Pixels
Global $dMaxMouseAngle = 135	; Degrees

Global $ptTurnLayerButtonDown[2]
Global $ptTurnLayerButtonUp[2]

Global $fTurn = False
Global $fTurnDone = 2

Global $fXYZlayer
Global $nCubeLayer
Global $fClockWise

Global $nDegrees = 0
Global $nMaxDegrees = 90
Global $nDeltaDegrees = 3 ; 3, 6, 9

Func TurnLayerButtonDown( ByRef $ptMouse )
	$ptTurnLayerButtonDown = $ptMouse
EndFunc

Func TurnLayerButtonUp( ByRef $ptMouse )
	Local $nCubeSide, $XYdirection
	$ptTurnLayerButtonUp = $ptMouse
	; When the mouse button is pressed, mouse moved, and button released
	; it's tested if the movement is able to rotate a layer in the cube.
	If TestTurnLayer( $iSize, $nCubeSide, $XYdirection ) Then
		SetTurnLayerVars( $nCubeSide, $XYdirection ) ; Set variables to turn the layer
		$nDegrees = $nDeltaDegrees	; Number of degrees to rotate the layer per step
		$fTurnDone = 0							; $fTurnDone is set to 2 when the rotation is done
		$fTurn = True								; $fTurn is set to True to turn the layer of the cube
		UpdateUndoLog()
		GUICtrlSetData( $cbLabel, $idxUndo )
		If $idxUndo = 1 Then _
			GUICtrlSetState( $cbUndo, $GUI_ENABLE )
		If $idxRedo Then
			GUICtrlSetState( $cbRedo, $GUI_DISABLE )
			$idxRedo = 0
			If $fLogsListsWin Then
				_GUICtrlListView_BeginUpdate( $hLVredo )
				_GUICtrlListView_DeleteAllItems( $hLVredo )
				_GUICtrlListView_EndUpdate( $hLVredo )
			EndIf
		EndIf
	EndIf
EndFunc

; Test if the mouse movement with the left mouse button pressed (or right if you have
; changed the buttons in the options) is able to rotate a layer in a 90 degrees turn.
Func TestTurnLayer( $iSize, ByRef $nCubeSide, ByRef $XYdirection )
	; Return False if a test isn't passed in this schedule:
	; 0. Test if the previous layer rotation is finished
	; 1. Test if the mouse movement is sufficiently large
	; 2. Is the mouse clicked inside the screen rectangle that includes the cube (a rough test)
	; 3. Test carefully which side of the cube is clicked (no rough test this time)
	; Now we know which side of the cube is clicked (the target of the mouse down point)
	; 4. Test if the mouse movement is approximately parallel to a pair of the edges of this side
	; 5. Test which layer on the side is clicked
	; Return True if all five tests are passed

	; 0. Test if the previous layer rotation is finished
	If $fTurnDone <> 2 Then Return False

	; The window coordinate system has (0,0) in upper left corner
	; The OpenGL viewport coordinate system has (0,0) in lower left corner
	; Adjust the mouse down/up points to the viewport coordinate system 
	Local $ptMouseDown[2], $ptMouseUp[2]
	$ptMouseDown[0] = $ptTurnLayerButtonDown[0]
	$ptMouseDown[1] = $iHeight - $ptTurnLayerButtonDown[1] - 1
	$ptMouseUp[0] = $ptTurnLayerButtonUp[0]
	$ptMouseUp[1] = $iHeight - $ptTurnLayerButtonUp[1] - 1

	; 1. Test if the mouse movement is sufficiently large
	; Return False if distance between down/up points is smaller than $dMinMouseMovement
	; A click where the mouse is moved a pixel or two between down/up points isn't considered as a turn
	Local $vMouse[2] ; A 2-dim vector in the XY-plane (screen)
	$vMouse[0] = $ptMouseDown[0] - $ptMouseUp[0]
	$vMouse[1] = $ptMouseDown[1] - $ptMouseUp[1]
	If VectorLength2( $vMouse ) < $dMinMouseMovement Then Return False

	; Project the corners of the cube into screen coordinates
	Local $vCubeCorners[8][3]	; 8 points in the XYZ-coordinate system
	Local $mProjection[16]		; 4 x 4 OpenGL projection matrix
	Local $nViewPort[4]				; x, y, w, h (w,h of the viewport equals w,h of the OpenGL GUI window)
	Local $d, $min, $max
	glGetFloatv( $GL_PROJECTION_MATRIX, $mProjection )
	glGetIntegerv( $GL_VIEWPORT, $nViewPort )
	$d = PieceDisplacement( $iSize )	; For Rubik's Cube $iSize = 3 and $d = -1
	$max = $d + $iSize - 1 + 0.5			; => $max =  1.5
	$min = $d - 0.5										; => $min = -1.5
	; Front corners
	; In object coordinates the cube has "center" in (0,0,0) and sides parallel to the XYZ-axes. gluProject() transforms the object coordi-
	; nates to window coordinates. Window coordinates have X-axis to the right, Y-axis upwards, Z-axis out of the window towards the viewer.
	gluProject( $max, $max, $max, $mModel, $mProjection, $nViewPort, $vCubeCorners[0][0], $vCubeCorners[0][1], $vCubeCorners[0][2] )
	gluProject( $max, $min, $max, $mModel, $mProjection, $nViewPort, $vCubeCorners[1][0], $vCubeCorners[1][1], $vCubeCorners[1][2] )
	gluProject( $min, $min, $max, $mModel, $mProjection, $nViewPort, $vCubeCorners[2][0], $vCubeCorners[2][1], $vCubeCorners[2][2] )
	gluProject( $min, $max, $max, $mModel, $mProjection, $nViewPort, $vCubeCorners[3][0], $vCubeCorners[3][1], $vCubeCorners[3][2] )
	; Back corners
	gluProject( $max, $max, $min, $mModel, $mProjection, $nViewPort, $vCubeCorners[4][0], $vCubeCorners[4][1], $vCubeCorners[4][2] )
	gluProject( $max, $min, $min, $mModel, $mProjection, $nViewPort, $vCubeCorners[5][0], $vCubeCorners[5][1], $vCubeCorners[5][2] )
	gluProject( $min, $min, $min, $mModel, $mProjection, $nViewPort, $vCubeCorners[6][0], $vCubeCorners[6][1], $vCubeCorners[6][2] )
	gluProject( $min, $max, $min, $mModel, $mProjection, $nViewPort, $vCubeCorners[7][0], $vCubeCorners[7][1], $vCubeCorners[7][2] )

	; 2. Is the mouse clicked inside the screen rectangle that includes the cube (a rough test)

	; Calculate the screen area (rectangle) that includes the entire cube
	; This calculation only involves the XY-coordinates
	Local $xMin = 10000, $xMax = -10000
	Local $yMin = 10000, $yMax = -10000
	For $i = 0 To 7
		If $vCubeCorners[$i][0] < $xMin Then $xMin = $vCubeCorners[$i][0]
		If $vCubeCorners[$i][0] > $xMax Then $xMax = $vCubeCorners[$i][0]
		If $vCubeCorners[$i][1] < $yMin Then $yMin = $vCubeCorners[$i][1]
		If $vCubeCorners[$i][1] > $yMax Then $yMax = $vCubeCorners[$i][1]
	Next

	; Return False if mouse wasn't clicked in this area.
	; To avoid calculations on a point which certainly is outside the cube.
	; This is a very rough test. A point can easily pass this test and still not be valid.
	; But if a point can't pass this test, it certainly isn't a valid point.
	If Not ( $xMin <= $ptMouseDown[0] And $ptMouseDown[0] <= $xMax And _
					 $yMin <= $ptMouseDown[1] And $ptMouseDown[1] <= $yMax ) Then Return False

	; 3. Test carefully which side of the cube is clicked (no rough test this time)

	; Create an array with the corners of the six sides
	Local $c = $vCubeCorners
	Local $vSideCorners[6][4][3] = [ _
		[ [$c[5][0],$c[5][1],$c[5][2]], [$c[1][0],$c[1][1],$c[1][2]], [$c[0][0],$c[0][1],$c[0][2]], [$c[4][0],$c[4][1],$c[4][2]] ], _ ; Right  = 0
		[ [$c[6][0],$c[6][1],$c[6][2]], [$c[2][0],$c[2][1],$c[2][2]], [$c[3][0],$c[3][1],$c[3][2]], [$c[7][0],$c[7][1],$c[7][2]] ], _ ; Left   = 1
		[ [$c[7][0],$c[7][1],$c[7][2]], [$c[4][0],$c[4][1],$c[4][2]], [$c[0][0],$c[0][1],$c[0][2]], [$c[3][0],$c[3][1],$c[3][2]] ], _ ; Top    = 2
		[ [$c[6][0],$c[6][1],$c[6][2]], [$c[5][0],$c[5][1],$c[5][2]], [$c[1][0],$c[1][1],$c[1][2]], [$c[2][0],$c[2][1],$c[2][2]] ], _ ; Bottom = 3
		[ [$c[2][0],$c[2][1],$c[2][2]], [$c[1][0],$c[1][1],$c[1][2]], [$c[0][0],$c[0][1],$c[0][2]], [$c[3][0],$c[3][1],$c[3][2]] ], _ ; Front  = 4
		[ [$c[6][0],$c[6][1],$c[6][2]], [$c[5][0],$c[5][1],$c[5][2]], [$c[4][0],$c[4][1],$c[4][2]], [$c[7][0],$c[7][1],$c[7][2]] ] ]  ; Back   = 5
	$c = 0

	; Test each of the six sides for a click
	Local $zMin = 10000, $zTmp, $nSide = -1
	For $side = 0 To 5
		If SideTest( $vSideCorners, $side, $ptMouseDown ) Then
			; In the XY-plane the four edges of a side creates a parallelogram (square with two pairs of oblique but parallel sides).
			; SideTest() tests if a point is inside this parallelogram. The test will normally match two sides. The Z-coordinates are
			; used to figure out the proper side.
			$zTmp = Min( Min( Min( $vSideCorners[$side][0][2], $vSideCorners[$side][1][2] ), $vSideCorners[$side][2][2] ), $vSideCorners[$side][3][2] )
			If $zTmp < $zMin Then
				$nSide = $side
				$zMin = $zTmp
			EndIf
		EndIf
	Next
	If $nSide = -1 Then Return False

	; 4. Test if the mouse movement is approximately parallel to a pair of the edges of this side

	; Calculate the angles between the mouse vector and the X- and Y-vectors
	; The X- and Y-vectors are parallel to the edges of the clicked side and they are vectors in the XY-plane
	Local $vX[2], $vY[2], $xAngle, $yAngle, $minAngle
	$vX[0] = $vSideCorners[$nSide][0][0] - $vSideCorners[$nSide][1][0]
	$vX[1] = $vSideCorners[$nSide][0][1] - $vSideCorners[$nSide][1][1]
	$vY[0] = $vSideCorners[$nSide][1][0] - $vSideCorners[$nSide][2][0]
	$vY[1] = $vSideCorners[$nSide][1][1] - $vSideCorners[$nSide][2][1]
	VectorNormalize2( $vMouse )
	VectorNormalize2( $vX )
	VectorNormalize2( $vY )
	$xAngle = UnitVectorAngle( $vX, $vMouse )
	$yAngle = UnitVectorAngle( $vY, $vMouse )

	; Calculate the minimum angle and add a small value (0.0025) to eliminate rounding errors.
	; Below $xAngle and $yAngle are compared to $minAngle. This is the reason for adding a small value.
	$minAngle = Min( Min( Min( Abs( $xAngle ), Abs( $yAngle ) ), Abs( $xAngle - 180 ) ), Abs( $yAngle - 180 ) ) + 0.0025

	; Return False if the minimum angle is greather than $dMaxMouseAngle
	; In that case the mouse movement hasn't been sufficiently parallel to the edges of the side
	If $minAngle > $dMaxMouseAngle Then Return False

	; 5. Test which layer on the side is clicked

	; Figure out which layer on the side is clicked.
	; Calculate a number $XYdirection to decide if the
	; mouse movement results in a positive or negative turn.
	; Positive turn = CCW turn, negative turn = CW turn.
	Local $nLayer, $corners[4], $aTmp[2]

	For $c = 0 To 3
		$aTmp[0] = $vSideCorners[$nSide][$c][0]
		$aTmp[1] = $vSideCorners[$nSide][$c][1]
		$corners[$c] = $aTmp
	Next

	If Abs( $xAngle ) <= $minAngle Then
		$XYdirection = 1
		$nLayer = XLayerTest( $corners, $ptMouseDown, $iSize )
	ElseIf Abs( $xAngle - 180 ) <= $minAngle Then
		$XYdirection = 2
		$nLayer = XLayerTest( $corners, $ptMouseDown, $iSize )
	ElseIf Abs( $yAngle ) <= $minAngle Then
		$XYdirection = 3
		$nLayer = YLayerTest( $corners, $ptMouseDown, $iSize )
	ElseIf Abs( $yAngle - 180 ) <= $minAngle Then
		$XYdirection = 4
		$nLayer = YLayerTest( $corners, $ptMouseDown, $iSize )
	Else
		Return False
	EndIf

	If $nLayer > -1 Then
		$nCubeSide = $nSide
		$nCubeLayer = $nLayer
		Return True
	Else
		Return False
	EndIf
EndFunc

; In the XY-plane the four edges of a side creates a parallelogram
; Parallelogram = square with two pairs of oblique but parallel sides
; SideTest() tests if a point is inside such a parallelogram
; Returns True if point is inside the side given by $side
; Right = 0, Left = 1, Top = 2, Bottom = 3, Front = 4, Back = 5
Func SideTest( ByRef $vSideCorners, $side, ByRef $point )
	Local $corners[4], $aTmp[2], $res1, $res2

	For $c = 0 To 3
		$aTmp[0] = $vSideCorners[$side][$c][0]
		$aTmp[1] = $vSideCorners[$side][$c][1]
		$corners[$c] = $aTmp
	Next

	; Return False if point is outside the first pair of parallel lines
	; The lines created by extending the first two parallel edges of the side
	$res1 = LineTest( $corners[0], $corners[1], $point )
	$res2 = LineTest( $corners[2], $corners[3], $point )
	If ( $res1 > 0 And $res2 > 0 ) Or ( $res1 < 0 And $res2 < 0 ) Then Return False

	; Return False if point is outside the last pair of parallel lines
	; The lines created by extending the last two parallel edges of the side
	$res1 = LineTest( $corners[0], $corners[3], $point )
	$res2 = LineTest( $corners[1], $corners[2], $point )
	If ( $res1 > 0 And $res2 > 0 ) Or ( $res1 < 0 And $res2 < 0 ) Then Return False

	Return True
EndFunc

; Returns  1 if $point is above/right of the line
;         -1 if $point is below/left of the line
;          0 if $point is on the line
Func LineTest( ByRef $LineStart, ByRef $LineEnd, ByRef $point )
	Local $dx = $LineEnd[0] - $LineStart[0], $xOnLine
	Local $dy = $LineEnd[1] - $LineStart[1], $yOnLine
	If Abs( $dx ) > Abs( $dy ) Then
    $yOnline = $dy/$dx * ($point[0] - $LineStart[0]) + $LineStart[1]
		If $point[1] > $yOnline Then
			Return 1
		ElseIf $point[1] < $yOnline Then
			Return -1
		Else
			Return 0
		EndIf
	Else
    $xOnline = $dx/$dy * ($point[1] - $LineStart[1]) + $LineStart[0]
		If $point[0] > $xOnline Then
			Return 1
		ElseIf $point[0] < $xOnline Then
			Return -1
		Else
			Return 0
		EndIf
	EndIf
EndFunc

; Figure out which layer on the side was clicked
; Use the same methods as in the side test
Func XLayerTest( $corners, $point, $iSize )
	Local $corner0 = $corners[0], $corner1 = $corners[1], $corner2 = $corners[2]
	Local $dx = ( $corner2[0] - $corner1[0] ) / $iSize, $dy = ( $corner2[1] - $corner1[1] ) / $iSize
	Local $LineStart1 = $corner0, $LineEnd1 = $corner1, $res1
	Local $LineStart2 = $corner0, $LineEnd2 = $corner1, $res2
	For $i = 0 To $iSize - 1
		$LineStart2[0] += $dx
		$LineEnd2[0] += $dx
		$LineStart2[1] += $dy
		$LineEnd2[1] += $dy
		$res1 = LineTest( $LineStart1, $LineEnd1, $point )
		$res2 = LineTest( $LineStart2, $LineEnd2, $point )
		If ( $res1 > 0 And $res2 < 0 ) Or _
			 ( $res1 < 0 And $res2 > 0 ) Then Return $i
		$LineStart1[0] += $dx
		$LineEnd1[0] += $dx
		$LineStart1[1] += $dy
		$LineEnd1[1] += $dy
	Next
	Return -1
EndFunc

; Figure out which layer on the side was clicked
; Use the same methods as in the side test
Func YLayerTest( $corners, $point, $iSize )
	Local $corner0 = $corners[0], $corner1 = $corners[1], $corner3 = $corners[3]
	Local $dx = ( $corner1[0] - $corner0[0] ) / $iSize, $dy = ( $corner1[1] - $corner0[1] ) / $iSize
	Local $LineStart1 = $corner0, $LineEnd1 = $corner3, $res1
	Local $LineStart2 = $corner0, $LineEnd2 = $corner3, $res2
	For $i = 0 To $iSize - 1
		$LineStart2[0] += $dx
		$LineEnd2[0] += $dx
		$LineStart2[1] += $dy
		$LineEnd2[1] += $dy
		$res1 = LineTest( $LineStart1, $LineEnd1, $point )
		$res2 = LineTest( $LineStart2, $LineEnd2, $point )
		If ( $res1 > 0 And $res2 < 0 ) Or _
			 ( $res1 < 0 And $res2 > 0 ) Then Return $i
		$LineStart1[0] += $dx
		$LineEnd1[0] += $dx
		$LineStart1[1] += $dy
		$LineEnd1[1] += $dy
	Next
	Return -1
EndFunc

; Set variables to turn a layer of the cube 90 degrees.
; Depending on the side clicked and $XYdirection it can
; be decided which layer to turn and the direction to turn.
Func SetTurnLayerVars( $nCubeSide, $XYdirection )
	Switch $XYdirection
		Case 1	; 0 <= $xAngle <= $dMaxMouseAngle
			Switch $nCubeSide
				Case $RIGHT, $BACK	; Turn Y-layer
					$fXYZlayer = 2
					$fClockWise = True
				Case $LEFT, $FRONT	; Turn Y-layer
					$fXYZlayer = 2
					$fClockWise = False
				Case $TOP						; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = True
				Case $BOTTOM				; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = False
			EndSwitch
		Case 2	; 180 <= $xAngle <= 180 + $dMaxMouseAngle
			Switch $nCubeSide
				Case $RIGHT, $BACK		; Turn Y-layer
					$fXYZlayer = 2
					$fClockWise = False
				Case $LEFT, $FRONT		; Turn Y-layer
					$fXYZlayer = 2
					$fClockWise = True
				Case $TOP							; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = False
				Case $BOTTOM					; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = True
			EndSwitch
		Case 3	; 0 <= $yAngle <= $dMaxMouseAngle
			Switch $nCubeSide
				Case $RIGHT						; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = False
				Case $LEFT						; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = True
				Case $TOP, $BACK			; Turn X-layer
					$fXYZlayer = 1
					$fClockWise = False
				Case $BOTTOM, $FRONT	; Turn X-layer
					$fXYZlayer = 1
					$fClockWise = True
			EndSwitch
		Case 4	; 180 <= $yAngle <= 180 + $dMaxMouseAngle
			Switch $nCubeSide
				Case $RIGHT						; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = True
				Case $LEFT						; Turn Z-layer
					$fXYZlayer = 3
					$fClockWise = False
				Case $TOP, $BACK			; Turn X-layer
					$fXYZlayer = 1
					$fClockWise = True
				Case $BOTTOM, $FRONT	; Turn X-layer
					$fXYZlayer = 1
					$fClockWise = False
			EndSwitch
	EndSwitch
EndFunc

; Turn a layer of the cube 90 degrees.
; The TurnLayer() function does the actual rotation of the pieces in the cube structure by
; changing the colors of the sides in the pieces and by reposition the pieces in the layer.
Func TurnLayer()
	Local $x0 = 0, $y0 = 0, $z0 = 0
	Local $x1 = $iSize - 1, $y1 = $iSize - 1, $z1 = $iSize - 1
	Local $xLayer = -1, $yLayer = -1, $zLayer = -1
	Local $aLayerPieces[$iSize][$iSize]
	Local $aPieceSides

	Switch $fXYZlayer
		Case 1 ; X
			$x0 = $nCubeLayer
			$x1 = $nCubeLayer
			$xLayer = $nCubeLayer
		Case 2 ; Y
			$y0 = $nCubeLayer
			$y1 = $nCubeLayer
			$yLayer = $nCubeLayer
		Case 3 ; Z
			$z0 = $nCubeLayer
			$z1 = $nCubeLayer
			$zLayer = $nCubeLayer
	EndSwitch

	; Rotate the pieces in the layer 90 degrees
	; Create new display lists for the pieces
	; Save the pieces in $aLayerPieces
	For $x = $x0 To $x1
		For $y = $y0 To $y1
			For $z = $z0 To $z1
				$aPieceSides = $aPieces[$x][$y][$z]
				If $aPieceSides[6] Then
					If $x = $xLayer Then
						TurnXpiece( $aPieceSides )
						CreatePiece( $aPieceSides )
						$aLayerPieces[$y][$z] = $aPieceSides
					ElseIf $y = $yLayer Then
						TurnYpiece( $aPieceSides )
						CreatePiece( $aPieceSides )
						$aLayerPieces[$x][$z] = $aPieceSides
					ElseIf $z = $zLayer Then
						TurnZpiece( $aPieceSides )
						CreatePiece( $aPieceSides )
						$aLayerPieces[$x][$y] = $aPieceSides
					EndIf
				EndIf
			Next
		Next
	Next

	; Rotate the layer in the cube structure.
	; When a layer is rotated 90 degrees, rows
	; will be columns and columns will be rows.
	Switch $fXYZlayer
		Case 1 ; X
			TurnXlayer( $aLayerPieces )
		Case 2 ; Y
			TurnYlayer( $aLayerPieces )
		Case 3 ; Z
			TurnZlayer( $aLayerPieces )
	EndSwitch

	; Put the pieces in $aLayerPieces back into the cube structure
	For $x = $x0 To $x1
		For $y = $y0 To $y1
			For $z = $z0 To $z1
				$aPieceSides = $aPieces[$x][$y][$z]
				If $aPieceSides[6] Then
					If $x = $xLayer Then
						$aPieces[$x][$y][$z] = $aLayerPieces[$y][$z]
					ElseIf $y = $yLayer Then
						$aPieces[$x][$y][$z] = $aLayerPieces[$x][$z]
					ElseIf $z = $zLayer Then
						$aPieces[$x][$y][$z] = $aLayerPieces[$x][$y]
					EndIf
				EndIf
			Next
		Next
	Next

	If $fLogsColorsWin Then
		Switch $fXYZlayer
			Case 1 ; X
				If $xLayer = 0 Then
					LogsColorsFillSquaresXside( $LEFT, $iSize )
				ElseIf $xLayer = $iSize - 1 Then
					LogsColorsFillSquaresXside( $RIGHT, $iSize )
				EndIf
				LogsColorsFillSquaresYside( $TOP,    $iSize, $xLayer )
				LogsColorsFillSquaresYside( $BOTTOM, $iSize, $xLayer )
				LogsColorsFillSquaresZside( $FRONT,  $iSize, $xLayer )
				LogsColorsFillSquaresZside( $BACK,   $iSize, $xLayer )
			Case 2 ; Y
				If $yLayer = 0 Then
					LogsColorsFillSquaresYside( $BOTTOM, $iSize )
				ElseIf $yLayer = $iSize - 1 Then
					LogsColorsFillSquaresYside( $TOP, $iSize )
				EndIf
				LogsColorsFillSquaresXside2( $RIGHT, $iSize, $yLayer )
				LogsColorsFillSquaresXside2( $LEFT,  $iSize, $yLayer )
				LogsColorsFillSquaresZside2( $FRONT, $iSize, $yLayer )
				LogsColorsFillSquaresZside2( $BACK,  $iSize, $yLayer )
			Case 3 ; Z
				If $zLayer = 0 Then
					LogsColorsFillSquaresZside( $BACK, $iSize )
				ElseIf $zLayer = $iSize - 1 Then
					LogsColorsFillSquaresZside( $FRONT, $iSize )
				EndIf
				LogsColorsFillSquaresXside ( $RIGHT,  $iSize, $zLayer )
				LogsColorsFillSquaresXside ( $LEFT,   $iSize, $zLayer )
				LogsColorsFillSquaresYside2( $TOP,    $iSize, $zLayer )
				LogsColorsFillSquaresYside2( $BOTTOM, $iSize, $zLayer )
		EndSwitch
	EndIf
EndFunc

; Rotate the pieces in the X-layer 90 degrees
Func TurnXpiece( ByRef $aPieceSides )
	Local $tmp
	If $fClockWise Then
		$tmp                  = $aPieceSides[$TOP]
		$aPieceSides[$TOP]    = $aPieceSides[$FRONT]
		$aPieceSides[$FRONT]  = $aPieceSides[$BOTTOM]
		$aPieceSides[$BOTTOM] = $aPieceSides[$BACK]
		$aPieceSides[$BACK]   = $tmp
	Else
		$tmp                  = $aPieceSides[$TOP]
		$aPieceSides[$TOP]    = $aPieceSides[$BACK]
		$aPieceSides[$BACK]   = $aPieceSides[$BOTTOM]
		$aPieceSides[$BOTTOM] = $aPieceSides[$FRONT]
		$aPieceSides[$FRONT]  = $tmp
	EndIf
EndFunc

; Rotate the pieces in the Y-layer 90 degrees
Func TurnYpiece( ByRef $aPieceSides )
	Local $tmp
	If $fClockWise Then
		$tmp                 = $aPieceSides[$FRONT]
		$aPieceSides[$FRONT] = $aPieceSides[$RIGHT]
		$aPieceSides[$RIGHT] = $aPieceSides[$BACK]
		$aPieceSides[$BACK]  = $aPieceSides[$LEFT]
		$aPieceSides[$LEFT]  = $tmp
	Else
		$tmp                 = $aPieceSides[$FRONT]
		$aPieceSides[$FRONT] = $aPieceSides[$LEFT]
		$aPieceSides[$LEFT]  = $aPieceSides[$BACK]
		$aPieceSides[$BACK]  = $aPieceSides[$RIGHT]
		$aPieceSides[$RIGHT] = $tmp
	EndIf
EndFunc

; Rotate the pieces in the Z-layer 90 degrees
Func TurnZpiece( ByRef $aPieceSides )
	Local $tmp
	If $fClockWise Then
		$tmp                  = $aPieceSides[$TOP]
		$aPieceSides[$TOP]    = $aPieceSides[$LEFT]
		$aPieceSides[$LEFT]   = $aPieceSides[$BOTTOM]
		$aPieceSides[$BOTTOM] = $aPieceSides[$RIGHT]
		$aPieceSides[$RIGHT]  = $tmp
	Else
		$tmp                  = $aPieceSides[$TOP]
		$aPieceSides[$TOP]    = $aPieceSides[$RIGHT]
		$aPieceSides[$RIGHT]  = $aPieceSides[$BOTTOM]
		$aPieceSides[$BOTTOM] = $aPieceSides[$LEFT]
		$aPieceSides[$LEFT]   = $tmp
	EndIf
EndFunc

; Rotate the X-layer in the cube structure.
; When a layer is rotated 90 degrees, rows
; will be columns and columns will be rows.
Func TurnXlayer( ByRef $aLayerPieces )
	Local $aTmp[$iSize][$iSize]
	Local $y1 = 0, $z1 = 0
	If $fClockWise Then
		For $z = 0 To $iSize - 1
			For $y = $iSize - 1 To 0 Step -1
				If IsArray( $aLayerPieces[$y][$z] ) Then _
					$aTmp[$y1][$z1] = $aLayerPieces[$y][$z]
				$z1 += 1
			Next
			$y1 += 1
			$z1 = 0
		Next
	Else
		For $z = $iSize - 1 To 0 Step -1
			For $y = 0 To $iSize - 1
				If IsArray( $aLayerPieces[$y][$z] ) Then _
					$aTmp[$y1][$z1] = $aLayerPieces[$y][$z]
				$z1 += 1
			Next
			$y1 += 1
			$z1 = 0
		Next
	EndIf
	$aLayerPieces = $aTmp
EndFunc

; Rotate the Y-layer in the cube structure.
; When a layer is rotated 90 degrees, rows
; will be columns and columns will be rows.
Func TurnYlayer( ByRef $aLayerPieces )
	Local $aTmp[$iSize][$iSize]
	Local $x1 = 0, $z1 = 0
	If $fClockWise Then
		For $z = $iSize - 1 To 0 Step -1
			For $x = 0 To $iSize - 1
				If IsArray( $aLayerPieces[$x][$z] ) Then _
					$aTmp[$x1][$z1] = $aLayerPieces[$x][$z]
				$z1 += 1
			Next
			$x1 += 1
			$z1 = 0
		Next
	Else
		For $z = 0 To $iSize - 1
			For $x = $iSize - 1 To 0 Step -1
				If IsArray( $aLayerPieces[$x][$z] ) Then _
					$aTmp[$x1][$z1] = $aLayerPieces[$x][$z]
				$z1 += 1
			Next
			$x1 += 1
			$z1 = 0
		Next
	EndIf
	$aLayerPieces = $aTmp
EndFunc

; Rotate the Z-layer in the cube structure.
; When a layer is rotated 90 degrees, rows
; will be columns and columns will be rows.
Func TurnZlayer( ByRef $aLayerPieces )
	Local $aTmp[$iSize][$iSize]
	Local $x1 = 0, $y1 = 0
	If $fClockWise Then
		For $y = 0 To $iSize - 1
			For $x = $iSize - 1 To 0 Step -1
				If IsArray( $aLayerPieces[$x][$y] ) Then _
					$aTmp[$x1][$y1] = $aLayerPieces[$x][$y]
				$y1 += 1
			Next
			$x1 += 1
			$y1 = 0
		Next
	Else
		For $y = $iSize - 1 To 0 Step -1
			For $x = 0 To $iSize - 1
				If IsArray( $aLayerPieces[$x][$y] ) Then _
					$aTmp[$x1][$y1] = $aLayerPieces[$x][$y]
				$y1 += 1
			Next
			$x1 += 1
			$y1 = 0
		Next
	EndIf
	$aLayerPieces = $aTmp
EndFunc

; Draws the cube by drawing one piece at a time
; Remark that the drawing is done by executing a precompiled display list
Func DrawTurn( $iSize, $nDeltaDegrees )
	Local $aPieceSides, $degrees
	Local $d = PieceDisplacement( $iSize )
	Local $xSpin = 0, $ySpin = 0, $zSpin = 0
	Local $xTurn = -1, $yTurn = -1, $zTurn = -1
	Switch $fXYZlayer
		Case 1 ; X
			$xSpin = 1
			$xTurn = $nCubeLayer
		Case 2 ; Y
			$ySpin = 1
			$yTurn = $nCubeLayer
		Case 3 ; Z
			$zSpin = 1
			$zTurn = $nCubeLayer
	EndSwitch
	If $fClockWise Then
		$degrees = -$nDegrees ; A clockwise turn is a turn in negative direction
	Else
		$degrees = $nDegrees
	EndIf
	For $x = 0 To $iSize - 1
		For $y = 0 To $iSize - 1
			For $z = 0 To $iSize - 1
				$aPieceSides = $aPieces[$x][$y][$z]
				If $aPieceSides[6] Then
					glPushMatrix()
						If $x = $xTurn Or $y = $yTurn Or $z = $zTurn Then _
							glRotatef( $degrees, $xSpin, $ySpin, $zSpin )
						glTranslatef( $d+$x, $d+$y, $d+$z )
						glCallList( $aPieceSides[6] )
					glPopMatrix()
				EndIf
			Next
		Next
	Next
	$nDegrees += $nDeltaDegrees
	If $nDegrees > $nMaxDegrees Then $fTurnDone = 1
EndFunc

; The DisplayTurn() function handles an entire OpenGL drawing cycle:
; Clears the window, saves the model view matrix, performs scaling and rotation, draws the cube, restores the matrix and swaps buffers.
; During this cycle there should not be any changes to the cube or the OpenGL matrices.
Func DisplayTurn( $iSize, $iAngX, $iAngY, $nDeltaDegrees )
	glClear( BitOR( $GL_COLOR_BUFFER_BIT, $GL_DEPTH_BUFFER_BIT ) )
	glPushMatrix()
		gluLookAt( 0,0,1+$dSphereDiameter/2, 0,0,0, 0,1,0 )
		glScalef( $dScale, $dScale, $dScale )
		glMultMatrixf( $mRotation )
		glRotatef( $iAngX, 1, 0, 0 )
		glRotatef( $iAngY, 0, 1, 0 )
		DrawTurn( $iSize, $nDeltaDegrees )
	glPopMatrix()
	SwapBuffers( $hDC )
EndFunc
