#include-once

Func PieceDisplacement( $iSize )
	If Mod( $iSize, 2 ) = 0 Then
		Return - $iSize / 2 + 0.5		; Even $iSize
	Else
		Return - ( $iSize - 1 ) / 2	; Odd $iSize
	EndIf
EndFunc

Func Min( $x, $y )
	If $x < $y Then
		Return $x
	Else
		Return $y
	EndiF
EndFunc

Func VectorLength2( ByRef $v )
	Return Sqrt( $v[0]*$v[0] + $v[1]*$v[1] )
EndFunc

Func VectorLength3( ByRef $v )
	Return Sqrt( $v[0]*$v[0] + $v[1]*$v[1] + $v[2]*$v[2] )
EndFunc

Func VectorNormalize2( ByRef $v )
	Local $l = VectorLength2( $v )
	$v[0] /= $l
	$v[1] /= $l
EndFunc

Func VectorNormalize3( ByRef $v )
	Local $l = VectorLength3( $v )
	$v[0] /= $l
	$v[1] /= $l
	$v[2] /= $l
EndFunc

; The important thing here is the vector dot product in relation to the law of cosines
; Two vectors a and b separated by angle v form a triangle with a third side c = a - b
; The vector dot product: c  c = a^2 + b^2 - 2ab
; The law of cosines: c^2 = a^2 + b^2 - 2ab*cos(v)
Func VectorDotProduct( ByRef $v1, ByRef $v2 )
	Return $v1[0]*$v2[0] + $v1[1]*$v2[1]
EndFunc

; The relation between the vector dot product and the law of cosines is used here
Func UnitVectorAngle( ByRef $v1, ByRef $v2 )
	Return ACos( VectorDotProduct( $v1, $v2 ) ) * ( 180 / $PI )
EndFunc

; The cross product of two vectors results in a vector which is
; perpendicular (normal) to the plane containing the two vectors.
Func VectorCrossProduct( ByRef $v1, ByRef $v2 )
	Local $v[3]
	$v[0] = $v1[1] * $v2[2] - $v1[2] * $v2[1]
	$v[1] = $v1[2] * $v2[0] - $v1[0] * $v2[2]
	$v[2] = $v1[0] * $v2[1] - $v1[1] * $v2[0]
	VectorNormalize3( $v )
	Return $v
EndFunc
